#' Performs a grid search over thresholds and returns accuracy statistics for a given numeric cue
#'
#' @param thresholds numeric. A vector of thresholds to consider
#' @param cue.v numeric. Feature values
#' @param criterion.v logical. Criterion values
#' @param directions character. Possible directions to consider
#' @param sens.w numeric.
#' @param cost.each numeric. Cost to add to each value (e.g.; cost of  the cue)
#' @param cost.outcomes list. A list of length 4 with names 'hi', 'fa', 'mi', and 'cr' specifying the costs of a hit, false alarm, miss, and correct rejection rspectively. E.g.; \code{cost.outcomes = listc("hi" = 0, "fa" = 10, "mi" = 20, "cr" = 0)} means that a false alarm and miss cost 10 and 20 respectively while correct decisions have no cost.
#' @param goal.threshold character. A string indicating the statistic to maximize when calculting cue thresholds: "acc" = overall accuracy, "wacc" = weighted accuracy, "bacc" = balanced accuracy
#'
threshold_numeric_grid <- function(thresholds,
                                   cue.v,
                                   criterion.v,
                                   directions = c(">", "<="),
                                   sens.w = .5,
                                   cost.each = 0,
                                   cost.outcomes = list(hi = 0, fa = 1, mi = 1, cr = 0),
                                   goal.threshold = "bacc") {

  # thresholds = cue_i_levels
  # cue.v = cue_i_v
  # criterion.v = criterion_v
  # sens.w = sens.w
  # directions = directions
  # cost.each = cue_i_cost
  # cost.outcomes = cost.outcomes
  # goal = goal

  thresholds_n <- length(thresholds)

  results_gt <- matrix(NA, nrow = thresholds_n, ncol = 5)

  # Loop over all thresholds
  # C++

  for(i in 1:thresholds_n) {

    threshold_i <- thresholds[i]

    # Create vector of decisions
    decisions_i <- cue.v > threshold_i

    # Calculate decisions
    hi_i <- sum(decisions_i == TRUE & criterion.v == TRUE)
    fa_i <- sum(decisions_i == TRUE & criterion.v == FALSE)
    mi_i <- sum(decisions_i == FALSE & criterion.v == TRUE)
    cr_i <- sum(decisions_i == FALSE & criterion.v == FALSE)

    n_i <- hi_i + fa_i + mi_i + cr_i

    # Return values to results
    results_gt[i, ] <- c(n_i, hi_i, fa_i, mi_i,  cr_i)

  }



  # Convert to named dataframe
  results_gt <- as.data.frame(results_gt)
  names(results_gt) <- c("n", "hi", "fa", "mi", "cr")
  results_gt$direction <- ">"
  results_gt$threshold <- thresholds

  # Get results if using <= threshold
  results_lt <- results_gt[,c(1, 4, 5, 2, 3)]
  names(results_lt) <- c("n", "hi", "fa", "mi", "cr")
  results_lt <- results_lt[,c("n", "hi", "fa", "mi", "cr")]



  results_lt$direction <- "<="
  results_lt$threshold <- thresholds

  if(setequal("<=", directions)) {

    results <- results_lt

  }

  if(setequal(">", directions)) {

    results <- results_gt

  }

  if(setequal(c(">", "<="), directions)) {

    results <- rbind(results_gt, results_lt)

  }


  new_stats <-  Add_Stats(results,
                          sens.w = sens.w,
                          cost.outcomes = cost.outcomes,
                          cost.each = cost.each)

  # Add accuracy statistics
  results <- cbind(results, new_stats)

  # Order by goal.threshold and change column order
  results <- results[order(-results[goal.threshold]), c("threshold", "direction", "n", "hi", "fa", "mi", "cr", "sens", "spec", "bacc", "acc", "wacc", "costout", "cost")]

  # Remove invalid directions
  results[results$direction %in% directions, ]

  return(results)

}
